import logging
import os
import subprocess
import tempfile
from pathlib import Path

from collagraph.cgx import cgx

from .template_formatter import format_template

logger = logging.getLogger(__name__)


def format_script(path, script_node, source_lines, check):
    start, end = script_node.location[0], script_node.end[0] - 1

    source = "".join(source_lines[start:end])

    ruff_command = ["ruff", "format"]
    if check:
        ruff_command.append("--check")

    # Write source to a temp file
    with tempfile.TemporaryDirectory() as directory:
        target_file = Path(directory) / "source.py"
        target_file.write_text(source)
        ruff_command.append(str(target_file))

        # Enable color output for ruff
        # TODO: check whether we want to force color or not?
        env = os.environ.copy()
        env["CLICOLOR_FORCE"] = "1"
        # Then run ruff to format the temp file
        output = subprocess.run(ruff_command, capture_output=True, text=True, env=env)
        with target_file.open(mode="r", encoding="utf-8") as fh:
            formatted_source = fh.readlines()

        stdout = output.stdout.replace(str(target_file), str(path))

    print(stdout, end="")  # noqa: T201
    if check:
        if output.returncode:
            exit(output.returncode)

    # Then return the contents
    return formatted_source, (start, end)


def format_file(path, check=False, write=True):
    """
    Format CGX files (the contents of the script tag) with ruff.

    Returns 0 if everything succeeded, or nothing changed.
    Returns 1 when running check and something would change.
    Returns list of lines when write is set to False instead
    of an error code (used for the test-suite).
    """
    path = Path(path)
    if path.suffix != ".cgx":
        return

    parser = cgx.CGXParser()
    parser.feed(path.read_text())

    with path.open(mode="r") as fh:
        lines = fh.readlines()

    script_node = parser.root.child_with_tag("script")
    template_node = parser.root.child_with_tag("template")

    script_content, script_location = format_script(path, script_node, lines, check)
    template_content, template_location = format_template(
        template_node, lines, parser=parser
    )

    changed = (
        lines[script_location[0] : script_location[1]] != script_content
        or lines[template_location[0] : template_location[1]] != template_content
    )
    if check:
        if changed:
            logger.warning(f"Would change: {path}")
            return 1
        return 0

    if script_location[0] > template_location[0]:
        lines[script_location[0] : script_location[1]] = script_content
        lines[template_location[0] : template_location[1]] = template_content
    else:
        lines[script_location[0] : script_location[1]] = script_content
        lines[template_location[0] : template_location[1]] = template_content
    if not write:
        return lines

    if changed:
        with path.open(mode="w") as fh:
            fh.writelines(lines)
    return 0



import ast
import os
import re
import subprocess
import tempfile
import textwrap
import tokenize
from pathlib import Path

from collagraph.cgx import cgx


def escape_ansi(line):
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def lint_file(path, fix=False, write=True):
    # plain_result = ast.unparse(tree)
    parser = cgx.CGXParser()
    parser.feed(Path(path).read_text())

    # Read the data from script block
    script_node = parser.root.child_with_tag("script")
    start, end = script_node.location[0], script_node.end[0] - 1
    script_range = range(start, end)

    tree, _ = cgx.construct_ast(path)

    # The only thing that we can't check is: RUF100: whether
    # there are unneeded noqa statement. That's because we use
    # these noqa statements to comment out the lines of the
    # virtual render function
    ruff_command = ["ruff", "check", "--ignore", "RUF100"]

    with path.open(mode="r", encoding="utf-8") as fh:
        source_raw = fh.readlines()

    # Comment out all non-script lines
    template_commented = [
        line if idx in script_range else "#\n" for idx, line in enumerate(source_raw)
    ]

    component_class_def = next(
        node for node in reversed(tree.body) if isinstance(node, ast.ClassDef)
    )
    component_class_name = component_class_def.name
    render_method = next(
        node
        for node in component_class_def.body
        if isinstance(node, ast.FunctionDef) and node.name == "render"
    )
    render_method = ast.unparse(render_method)
    render_method = textwrap.indent(render_method, "    ")

    # Insert a line that mimics a subclass so that any code that follows the last
    # class definition, won't break the virtual render method
    template_commented.append(
        f"class Virtual{component_class_name}({component_class_name}):\n"
    )
    # Append the virtual render method into the virtual subclass definition
    template_commented.extend(
        [f"{line}  # noqa\n" for line in render_method.split("\n")]
    )

    with tempfile.TemporaryDirectory() as directory:
        target_file = Path(directory) / "source.py"
        target_file.write_text("".join(template_commented))

        ruff_command.append(str(target_file))

        # Force color
        # TODO: figure out if we can detect if we can use color or not
        env = os.environ.copy()
        env["CLICOLOR_FORCE"] = "1"

        result = subprocess.run(
            ruff_command,
            capture_output=True,
            text=True,
            env=env,
        )

    stdout = result.stdout.replace(str(target_file), str(path))
    stdout.strip()
    print(stdout, end="")  # noqa: T201
    return result.returncode


def read_lines_from_filename_patched(path: Path) -> list[str]:
    """Read the lines for a file."""
    try:
        parser = cgx.CGXParser()
        parser.feed(Path(path).read_text())

        # Read the data from script block
        script_node = parser.root.child_with_tag("script")
        start, end = script_node.location[0], script_node.end[0] - 1

        with tokenize.open(path) as fh:
            result = fh.readlines()

            actual_result = result[start:end]
            # Prepend some empty (commented) lines to make the errors
            # point out the right location in the cgx file
            actual_result = [
                # prepend with some empty lines
                *(start * ["# noqa\n"]),
                *actual_result,
                # append some more empty lines
                *(9 * ["# noqa\n"]),
            ]

            return actual_result

    except (SyntaxError, UnicodeError):
        # If we can't detect the codec with tokenize.detect_encoding, or
        # the detected encoding is incorrect, just fallback to latin-1.
        with open(path, encoding="latin-1") as fd:
            return fd.readlines()



import pytest

from ruff_cgx.__main__ import main


def test_check_command(capsys, data_path):
    simple_cgx = data_path / "lint.cgx"

    with pytest.raises(SystemExit) as e:
        main(["check", str(simple_cgx)])

    captured = capsys.readouterr()

    assert e.value.code == 1
    assert "tests/data/lint.cgx:72:27" in captured.out
    assert "F401" in captured.out
    assert "`PySide6.QtGui.QAction` imported but unused" in captured.out
    assert "Found 1 error." in captured.out
